# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# detector_bayesian_torch.py
# Description: Implementation of Bayesian detector for SynthID algorithm in PyTorch
# ============================================

import abc
from collections.abc import Mapping, Sequence
import enum
import functools
import gc
from typing import Any
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from sklearn import model_selection
import tqdm


def pad_to_len(
    arr: torch.Tensor,
    target_len: int,
    *,
    left_pad: bool,
    eos_token: int,
    device: torch.device,
) -> torch.Tensor:
    """Pad or truncate array to given length."""
    if arr.shape[1] < target_len:
        shape_for_ones = list(arr.shape)
        shape_for_ones[1] = target_len - shape_for_ones[1]
        padded = torch.ones(
            shape_for_ones,
            device=device,
            dtype=torch.long,
        ) * eos_token
        
        if not left_pad:
            return torch.cat((arr, padded), dim=1)
        else:
            return torch.cat((padded, arr), dim=1)
    else:
        return arr[:, :target_len]


def filter_and_truncate(
    outputs: torch.Tensor,
    truncation_length,
    eos_token_mask: torch.Tensor, 
) -> torch.Tensor:
    """Filter and truncate outputs to given length."""
    if truncation_length:
        outputs = outputs[:, :truncation_length]
        truncation_mask = torch.sum(eos_token_mask, dim=1) >= truncation_length
        return outputs[truncation_mask, :]
    return outputs


def process_outputs_for_training(
    all_outputs: Sequence[torch.Tensor],
    logits_processor,
    tokenizer: Any,
    *,
    pos_truncation_length,
    neg_truncation_length, 
    max_length: int,
    is_cv: bool,
    is_pos: bool,
    torch_device: torch.device,
) -> tuple[Sequence[torch.Tensor], Sequence[torch.Tensor]]:
    """Process raw model outputs for training."""
    all_masks = []
    all_g_values = []
    
    for outputs in tqdm.tqdm(all_outputs):
        eos_token_mask = logits_processor.compute_eos_token_mask(
            input_ids=outputs,
            eos_token_id=tokenizer.eos_token_id,
        )
        
        if is_pos or is_cv:
            outputs = filter_and_truncate(
                outputs, pos_truncation_length, eos_token_mask
            )
        elif not is_pos and not is_cv:
            outputs = filter_and_truncate(
                outputs, neg_truncation_length, eos_token_mask
            )
            
        if outputs.shape[0] == 0:
            continue
            
        outputs = pad_to_len(
            outputs,
            max_length,
            left_pad=False, 
            eos_token=tokenizer.eos_token_id,
            device=torch_device,
        )

        eos_token_mask = logits_processor.compute_eos_token_mask(
            input_ids=outputs,
            eos_token_id=tokenizer.eos_token_id,
        )

        context_repetition_mask = logits_processor.compute_context_repetition_mask(
            input_ids=outputs,
        )

        context_repetition_mask = pad_to_len(
            context_repetition_mask,
            max_length,
            left_pad=True,
            eos_token=0,
            device=torch_device,
        )
        
        combined_mask = context_repetition_mask * eos_token_mask

        g_values = logits_processor.compute_g_values(
            input_ids=outputs,
        )

        g_values = pad_to_len(
            g_values, 
            max_length,
            left_pad=True,
            eos_token=0,
            device=torch_device
        )

        all_masks.append(combined_mask)
        all_g_values.append(g_values)
        
    return all_masks, all_g_values


class LikelihoodModel(nn.Module, abc.ABC):
    """Base class for likelihood models."""

    @abc.abstractmethod
    def forward(self, g_values: torch.Tensor) -> torch.Tensor:
        """Compute likelihoods given g-values."""
        pass


class LikelihoodModelWatermarked(LikelihoodModel):
    """Model for P(g_values|watermarked)."""

    def __init__(self, watermarking_depth: int):
        super().__init__()
        self.watermarking_depth = watermarking_depth
        
        # Initialize parameters
        self.beta = nn.Parameter(
            -2.5 + 0.001 * torch.randn(1, 1, watermarking_depth)
        )
        self.delta = nn.Parameter(
            0.001 * torch.randn(1, 1, watermarking_depth, watermarking_depth)
        )

    def l2_loss(self) -> torch.Tensor:
        return torch.sum(self.delta ** 2)

    def _compute_latents(
        self, g_values: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Compute latent probabilities."""
        x = g_values.unsqueeze(-2).repeat(
            1, 1, self.watermarking_depth, 1
        )
        x = torch.tril(x, diagonal=-1)
        
        logits = torch.sum(self.delta * x, dim=-1) + self.beta
        
        p_two_unique_tokens = torch.sigmoid(logits)
        p_one_unique_token = 1 - p_two_unique_tokens
        
        return p_one_unique_token, p_two_unique_tokens

    def forward(self, g_values: torch.Tensor) -> torch.Tensor:
        """Compute P(g_values|watermarked)."""
        p_one_unique_token, p_two_unique_tokens = self._compute_latents(g_values)
        return 0.5 * ((g_values + 0.5) * p_two_unique_tokens + p_one_unique_token)


class LikelihoodModelUnwatermarked(LikelihoodModel):
    """Model for P(g_values|unwatermarked)."""

    def forward(self, g_values: torch.Tensor) -> torch.Tensor:
        """Compute P(g_values|unwatermarked)."""
        return 0.5 * torch.ones_like(g_values)


class BayesianDetectorModule(nn.Module):
    """Bayesian detector model."""

    def __init__(
        self,
        watermarking_depth: int,
        baserate: float = 0.5
    ):
        super().__init__()
        self.watermarking_depth = watermarking_depth
        self.baserate = baserate
        
        self.likelihood_model_watermarked = LikelihoodModelWatermarked(
            watermarking_depth=watermarking_depth
        )
        self.likelihood_model_unwatermarked = LikelihoodModelUnwatermarked()
        self.prior = nn.Parameter(torch.tensor([baserate]))

    def l2_loss(self) -> torch.Tensor:
        return self.likelihood_model_watermarked.l2_loss()

    def forward(
        self,
        g_values: torch.Tensor,
        mask: torch.Tensor,
    ) -> torch.Tensor:
        """Compute P(watermarked|g_values)."""
        likelihoods_watermarked = self.likelihood_model_watermarked(g_values)
        likelihoods_unwatermarked = self.likelihood_model_unwatermarked(g_values)
        
        mask = mask.unsqueeze(-1)
        prior = torch.clamp(self.prior, 1e-5, 1 - 1e-5)
        
        log_likelihoods_watermarked = torch.log(
            torch.clamp(likelihoods_watermarked, min=1e-30)
        )
        log_likelihoods_unwatermarked = torch.log(
            torch.clamp(likelihoods_unwatermarked, min=1e-30)
        )
        
        log_odds = log_likelihoods_watermarked - log_likelihoods_unwatermarked
        
        relative_surprisal_likelihood = torch.sum(
            log_odds * mask, dim=(1, 2)
        )
        
        relative_surprisal_prior = torch.log(prior) - torch.log(1 - prior)
        relative_surprisal = relative_surprisal_prior + relative_surprisal_likelihood
        
        return torch.sigmoid(relative_surprisal)


def train_model(
    detector_module: BayesianDetectorModule,
    g_values: torch.Tensor,
    mask: torch.Tensor,
    watermarked: torch.Tensor,
    epochs: int = 250,
    learning_rate: float = 1e-3,
    minibatch_size: int = 64,
    l2_weight: float = 0.0,
    g_values_val: torch.Tensor = None,
    mask_val: torch.Tensor = None,
    watermarked_val: torch.Tensor = None,
    verbose: bool = False,
) -> tuple[dict, float]:
    """Train the detector model."""
    
    optimizer = torch.optim.Adam(detector_module.parameters(), lr=learning_rate)
    
    history = {}
    min_val_loss = float('inf')
    best_state = None
    
    for epoch in range(epochs):
        detector_module.train()
        
        # Training
        train_losses = []
        for i in range(0, len(g_values), minibatch_size):
            batch_g = g_values[i:i+minibatch_size]
            batch_m = mask[i:i+minibatch_size] 
            batch_w = watermarked[i:i+minibatch_size]

            optimizer.zero_grad()
            
            pred = detector_module(batch_g, batch_m)
            loss = F.binary_cross_entropy(
                pred, batch_w.float()
            ) + l2_weight * detector_module.l2_loss()
            
            train_losses.append(loss.item())
            
            loss.backward()
            optimizer.step()
            
        avg_train_loss = sum(train_losses) / len(train_losses)
            
        # Validation
        detector_module.eval()
        with torch.no_grad():
            if g_values_val is not None:
                val_pred = detector_module(g_values_val, mask_val)
                val_loss = F.binary_cross_entropy(
                    val_pred, watermarked_val.float()
                )
                
                if val_loss < min_val_loss:
                    min_val_loss = val_loss
                    best_state = detector_module.state_dict()
                    
                if verbose:
                    print(f"Epoch {epoch}: train_loss {avg_train_loss:.4f}, val_loss {val_loss:.4f}")
                    
            history[epoch] = {
                'train_loss': avg_train_loss,
                'val_loss': val_loss.item() if g_values_val is not None else None
            }
            
    if best_state is not None:
        detector_module.load_state_dict(best_state)
    return history, min_val_loss


class RawBayesianDetector:
    """Bayesian detector for watermark detection."""

    def __init__(
        self,
        logits_processor,
        tokenizer: Any,
        detector_module: BayesianDetectorModule,
    ):
        self.detector_module = detector_module
        self.logits_processor = logits_processor
        self.tokenizer = tokenizer

    def score(self, outputs: torch.Tensor) -> torch.Tensor:
        """Score outputs for watermark detection."""
        eos_token_mask = self.logits_processor.compute_eos_token_mask(
            input_ids=outputs,
            eos_token_id=self.tokenizer.eos_token_id,
        )[:, self.logits_processor.ngram_len - 1:]

        context_repetition_mask = self.logits_processor.compute_context_repetition_mask(
            input_ids=outputs,
        )

        combined_mask = context_repetition_mask * eos_token_mask

        g_values = self.logits_processor.compute_g_values(
            input_ids=outputs,
        )

        with torch.no_grad():
            return self.detector_module(g_values, combined_mask)

    @classmethod
    def train_best_detector(
        cls,
        *,
        tokenized_wm_outputs,
        tokenized_uwm_outputs,
        logits_processor,
        tokenizer: Any,
        torch_device: torch.device,
        test_size: float = 0.3,
        pos_truncation_length: int = 200, 
        neg_truncation_length: int = 100,
        max_padded_length: int = 2300,
        n_epochs: int = 50,
        learning_rate: float = 2.1e-2,
        l2_weights: np.ndarray = np.logspace(-3, -2, num=4),
        verbose: bool = False,
    ) -> tuple["RawBayesianDetector", float]:
        """Train the best detector model."""
        
        if torch_device.type not in ("cuda", "tpu"):
            raise ValueError(
                "Training is unstable on CPUs. Please use GPU or TPU."
            )

        # Process outputs
        (
            train_g_values,
            train_masks,
            train_labels,
            cv_g_values,
            cv_masks, 
            cv_labels,
        ) = cls.process_raw_model_outputs(
            tokenized_wm_outputs=tokenized_wm_outputs,
            tokenized_uwm_outputs=tokenized_uwm_outputs,
            logits_processor=logits_processor,
            tokenizer=tokenizer,
            torch_device=torch_device,
            test_size=test_size,
            pos_truncation_length=pos_truncation_length,
            neg_truncation_length=neg_truncation_length,
            max_padded_length=max_padded_length,
        )

        # Convert numpy arrays to torch tensors
        train_g_values = torch.from_numpy(train_g_values).to(torch_device)
        train_masks = torch.from_numpy(train_masks).to(torch_device)
        train_labels = torch.from_numpy(train_labels).to(torch_device)
        cv_g_values = torch.from_numpy(cv_g_values).to(torch_device)
        cv_masks = torch.from_numpy(cv_masks).to(torch_device)
        cv_labels = torch.from_numpy(cv_labels).to(torch_device)

        # Train models with different L2 weights
        best_detector = None
        lowest_loss = float('inf')
        
        for l2_weight in l2_weights:
            detector_module = BayesianDetectorModule(
                watermarking_depth=len(logits_processor.keys)
            ).to(torch_device)
            
            _, min_val_loss = train_model(
                detector_module=detector_module,
                g_values=train_g_values,
                mask=train_masks,
                watermarked=train_labels,
                g_values_val=cv_g_values,
                mask_val=cv_masks,
                watermarked_val=cv_labels,
                learning_rate=learning_rate,
                l2_weight=l2_weight,
                epochs=n_epochs,
                verbose=verbose,
            )

            if min_val_loss < lowest_loss:
                lowest_loss = min_val_loss
                best_detector = detector_module

        return cls(logits_processor, tokenizer, best_detector), lowest_loss

    @classmethod
    def process_raw_model_outputs(
        cls,
        *,
        tokenized_wm_outputs,
        tokenized_uwm_outputs,
        logits_processor,
        tokenizer: Any,
        torch_device: torch.device,
        test_size: float = 0.3,
        pos_truncation_length: int = 200,
        neg_truncation_length: int = 100,
        max_padded_length: int = 2300,
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Process raw models outputs into inputs for training.

        Args:
            tokenized_wm_outputs: tokenized outputs of watermarked data
            tokenized_uwm_outputs: tokenized outputs of unwatermarked data
            logits_processor: logits processor used for watermarking
            tokenizer: tokenizer used for the model
            torch_device: torch device to use
            test_size: test size for train-test split
            pos_truncation_length: Length to truncate wm outputs
            neg_truncation_length: Length to truncate uwm outputs
            max_padded_length: Length to pad truncated outputs

        Returns:
            Tuple of (train_g_values, train_masks, train_labels, 
                     cv_g_values, cv_masks, cv_labels)
        """
        # Split data into train and CV
        train_wm_outputs, cv_wm_outputs = model_selection.train_test_split(
            tokenized_wm_outputs, test_size=test_size
        )

        train_uwm_outputs, cv_uwm_outputs = model_selection.train_test_split(
            tokenized_uwm_outputs, test_size=test_size
        )

        # Process both train and CV data
        wm_masks_train, wm_g_values_train = process_outputs_for_training(
            [torch.tensor(outputs, device=torch_device, dtype=torch.long)
             for outputs in train_wm_outputs],
            logits_processor=logits_processor,
            tokenizer=tokenizer,
            pos_truncation_length=pos_truncation_length,
            neg_truncation_length=neg_truncation_length,
            max_length=max_padded_length,
            is_pos=True,
            is_cv=False,
            torch_device=torch_device,
        )

        wm_masks_cv, wm_g_values_cv = process_outputs_for_training(
            [torch.tensor(outputs, device=torch_device, dtype=torch.long)
             for outputs in cv_wm_outputs],
            logits_processor=logits_processor,
            tokenizer=tokenizer,
            pos_truncation_length=pos_truncation_length,
            neg_truncation_length=neg_truncation_length,
            max_length=max_padded_length,
            is_pos=True,
            is_cv=True,
            torch_device=torch_device,
        )

        uwm_masks_train, uwm_g_values_train = process_outputs_for_training(
            [torch.tensor(outputs, device=torch_device, dtype=torch.long)
             for outputs in train_uwm_outputs],
            logits_processor=logits_processor,
            tokenizer=tokenizer,
            pos_truncation_length=pos_truncation_length,
            neg_truncation_length=neg_truncation_length,
            max_length=max_padded_length,
            is_pos=False,
            is_cv=False,
            torch_device=torch_device,
        )

        uwm_masks_cv, uwm_g_values_cv = process_outputs_for_training(
            [torch.tensor(outputs, device=torch_device, dtype=torch.long)
             for outputs in cv_uwm_outputs],
            logits_processor=logits_processor,
            tokenizer=tokenizer,
            pos_truncation_length=pos_truncation_length,
            neg_truncation_length=neg_truncation_length,
            max_length=max_padded_length,
            is_pos=False,
            is_cv=True,
            torch_device=torch_device,
        )

        # Concatenate data
        wm_masks_train = torch.cat(wm_masks_train, dim=0)
        wm_g_values_train = torch.cat(wm_g_values_train, dim=0)
        wm_labels_train = torch.ones((wm_masks_train.shape[0],), dtype=torch.bool)
        
        wm_masks_cv = torch.cat(wm_masks_cv, dim=0)
        wm_g_values_cv = torch.cat(wm_g_values_cv, dim=0)
        wm_labels_cv = torch.ones((wm_masks_cv.shape[0],), dtype=torch.bool)

        uwm_masks_train = torch.cat(uwm_masks_train, dim=0)
        uwm_g_values_train = torch.cat(uwm_g_values_train, dim=0)
        uwm_labels_train = torch.zeros((uwm_masks_train.shape[0],), dtype=torch.bool)
        
        uwm_masks_cv = torch.cat(uwm_masks_cv, dim=0)
        uwm_g_values_cv = torch.cat(uwm_g_values_cv, dim=0)
        uwm_labels_cv = torch.zeros((uwm_masks_cv.shape[0],), dtype=torch.bool)

        # Combine positive and negative examples
        train_g_values = torch.cat((wm_g_values_train, uwm_g_values_train), dim=0).cpu().numpy()
        train_masks = torch.cat((wm_masks_train, uwm_masks_train), dim=0).cpu().numpy()
        train_labels = torch.cat((wm_labels_train, uwm_labels_train), dim=0).cpu().numpy()

        cv_g_values = torch.cat((wm_g_values_cv, uwm_g_values_cv), dim=0).cpu().numpy()
        cv_masks = torch.cat((wm_masks_cv, uwm_masks_cv), dim=0).cpu().numpy()
        cv_labels = torch.cat((wm_labels_cv, uwm_labels_cv), dim=0).cpu().numpy()

        # Free GPU memory
        del (wm_g_values_train, wm_labels_train, wm_masks_train,
             wm_g_values_cv, wm_labels_cv, wm_masks_cv)
        gc.collect()
        torch.cuda.empty_cache()

        # Shuffle data
        train_g_values = np.squeeze(train_g_values)
        train_masks = np.squeeze(train_masks)
        train_labels = np.squeeze(train_labels)

        cv_g_values = np.squeeze(cv_g_values)
        cv_masks = np.squeeze(cv_masks)
        cv_labels = np.squeeze(cv_labels)

        # Shuffle training data
        train_idx = np.random.permutation(len(train_labels))
        train_g_values = train_g_values[train_idx]
        train_masks = train_masks[train_idx]
        train_labels = train_labels[train_idx]

        # Shuffle CV data
        cv_idx = np.random.permutation(len(cv_labels))
        cv_g_values = cv_g_values[cv_idx]
        cv_masks = cv_masks[cv_idx]
        cv_labels = cv_labels[cv_idx]

        return (train_g_values, train_masks, train_labels,
                cv_g_values, cv_masks, cv_labels)


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# =============================================
# openai_utils.py
# Description: Utility functions for OpenAI API
# =============================================
import os
import time
import openai
from exceptions.exceptions import OpenAIModelConfigurationError


class OpenAIAPI:
    """API class for OpenAI API."""
    def __init__(self, model, temperature, system_content):
        """
            Initialize OpenAI API with model, temperature, and system content.

            Parameters:
                model (str): Model name for OpenAI API.
                temperature (float): Temperature value for OpenAI API.
                system_content (str): System content for OpenAI API.
        """

        self.model = model
        self.temperature = temperature
        self.system_content = system_content
        self.client = openai.OpenAI()
        

        # List of supported models
        supported_models = ['gpt-3.5-turbo', 'gpt-4']

        # Check if the provided model is within the supported models
        if self.model not in supported_models:
            raise OpenAIModelConfigurationError(f"Unsupported model '{self.model}'. Supported models are {supported_models}.")

    def get_result_from_gpt4(self, query):
        """get result from GPT-4 model."""
        response = self.client.chat.completions.create(
            model="gpt-4-0613",
            temperature=self.temperature,
            messages=[
                {"role": "system", "content": self.system_content},
                {"role": "user", "content": query},
            ]
        )
        return response
    
    def get_result_from_gpt3_5(self, query):
        """get result from GPT-3.5 model."""
        response = self.client.chat.completions.create(
            model="gpt-3.5-turbo",
            temperature=self.temperature,
            messages=[
                {"role": "system", "content": self.system_content},
                {"role": "user", "content": query},
            ]
        )
        return response

    def get_result(self, query):
        """get result from OpenAI API."""
        while True:
            try:
                if self.model == 'gpt-3.5-turbo':
                    result = self.get_result_from_gpt3_5(query)
                elif self.model == 'gpt-4':
                    result = self.get_result_from_gpt4(query)
                break
            except Exception as e:
                print(f"OpenAI API error: {str(e)}")
            time.sleep(10)
        return result.choices[0].message.content


if __name__ == "__main__":
    openai_util = OpenAIAPI('gpt-3.5-turbo', 0.2, "Your are a helpful assistant to rewrite the text.")
    text =  "Rewrite the following paragraph:\n "
    text += "Whoever gets him, they'll be getting a good one,\" David Montgomery said. INDIANAPOLIS \u2014 Hakeem Butler has been surrounded by some of the best wide receivers on the planet this month, so it's only natural he'd get a touch of star-struck when he got a chance to meet Randy Moss, a Hall of Famer and someone Butler idolized growing up and still considers an inspiration to this day.\n\"It's crazy to see someone you've watched on TV and seen on YouTube and then be standing right there and you're like, 'Man, he's just a regular dude,'\" Butler, a 6-foot 5, 227-pound receiver from Iowa State, said Thursday at the NFL Scouts' Luncheon, which kicks off the week of the NFL Scouts' Breakfast and the NFL Scouts' Luncheon at the JW Marriott Indianapolis, which takes place Sunday and Monday, Feb. 1 and 2, at Lucas Oil Stadium, 1 Lucas Oil Stadium Dr., Indianapolis, IN 46225, and which is open to all NFL Scouts and team personnel and others with a pass from the NFL Scouts' Breakfast and the NFL Scouts' Luncheon, and which is sponsored by the NFL Scouts' Breakfast and the NFL Scouts' Luncheon, and which is organized by the NFL Scouts' Breakfast and the NFL Scouts' Luncheon, and which is hosted by the"
    result = openai_util.get_result(text)
    print(result)


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ===================================================
# mersenne.py
# Description: Implementation of Mersenne Twister RNG
# ===================================================

class MersenneRNG(object):
    """Mersenne Twister RNG implementation."""

    def __init__(self, seed: int = 42):
        """
            Initialize the Mersenne Twister RNG.

            Parameters:
                seed (int): Seed for the RNG.
        """
        self.state = [0] * 624
        self.f = 1812433253
        self.m = 397
        self.u = 11
        self.s = 7
        self.b = 0x9D2C5680
        self.t = 15
        self.c = 0xEFC60000
        self.l = 18
        self.index = 624
        self.lower_mask = (1 << 31) - 1
        self.upper_mask = 1 << 31

        # update state
        self.state[0] = seed
        for i in range(1, 624):
            self.state[i] = self.int_32(self.f * (self.state[i - 1] ^ (self.state[i - 1] >> 30)) + i)

    def twist(self) -> None:
        """Twist the RNG."""
        for i in range(624):
            temp = self.int_32((self.state[i] & self.upper_mask) + (self.state[(i + 1) % 624] & self.lower_mask))
            temp_shift = temp >> 1
            if temp % 2 != 0:
                temp_shift = temp_shift ^ 0x9908b0df
            self.state[i] = self.state[(i + self.m) % 624] ^ temp_shift
        self.index = 0

    def int_32(self, number: int) -> int:
        """Return 32-bit integer."""
        return int(0xFFFFFFFF & number)

    def randint(self) -> int:
        """Return random integer."""

        if self.index >= 624:
            self.twist()
        y = self.state[self.index]
        y = y ^ (y >> self.u)
        y = y ^ ((y << self.s) & self.b)
        y = y ^ ((y << self.t) & self.c)
        y = y ^ (y >> self.l)
        self.index += 1
        return self.int_32(y)

    def rand(self) -> float:
        """Return random float."""
        return self.randint() * (1.0 / 4294967296.0)

    def randperm(self, n: int) -> list[int]:
        """Return random permutation of n elements."""
        
        # Fisher-Yates shuffle
        p = list(range(n))
        for i in range(n - 1, 0, -1):
            j = self.randint() % i
            p[i], p[j] = p[j], p[i]

        return p



# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# kgw.py
# Description: Implementation of KGW algorithm
# ============================================

import torch
from math import sqrt
from functools import partial
from ..base import BaseWatermark, BaseConfig
from watermark.utils import load_config_file
from watermark.transformers_config import TransformersConfig
# from exceptions.exceptions import AlgorithmNameMismatchError
from transformers import LogitsProcessor, LogitsProcessorList
# from visualize.data_for_visualization import DataForVisualization


class KGWConfig(BaseConfig):
    """Config class for KGW algorithm"""
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.gamma = self.config_dict['gamma']
        self.delta = self.config_dict['delta']
        self.hash_key = self.config_dict['hash_key']
        self.z_threshold = self.config_dict['z_threshold']
        self.prefix_length = self.config_dict['prefix_length']
        self.f_scheme = self.config_dict['f_scheme']
        self.window_scheme = self.config_dict['window_scheme']
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'KGW'

class KGWUtils:
    """Utility class for KGW algorithm, contains helper functions."""

    def __init__(self, config: KGWConfig, *args, **kwargs) -> None:
        """
            Initialize the KGW utility class.

            Parameters:
                config (KGWConfig): Configuration for the KGW algorithm.
        """
        self.config = config
        self.rng = torch.Generator(device=self.config.device)
        self.rng.manual_seed(self.config.hash_key)
        self.prf = torch.randperm(self.config.vocab_size, device=self.config.device, generator=self.rng)
        self.f_scheme_map = {"time": self._f_time, "additive": self._f_additive, "skip": self._f_skip, "min": self._f_min}
        self.window_scheme_map = {"left": self._get_greenlist_ids_left, "self": self._get_greenlist_ids_self}

    def _f(self, input_ids: torch.LongTensor) -> int:
        """Get the previous token."""
        return int(self.f_scheme_map[self.config.f_scheme](input_ids))
    
    def _f_time(self, input_ids: torch.LongTensor):
        """Get the previous token time."""
        time_result = 1
        for i in range(0, self.config.prefix_length):
            time_result *= input_ids[-1 - i].item()
        return self.prf[time_result % self.config.vocab_size]
    
    def _f_additive(self, input_ids: torch.LongTensor):
        """Get the previous token additive."""
        additive_result = 0
        for i in range(0, self.config.prefix_length):
            additive_result += input_ids[-1 - i].item()
        return self.prf[additive_result % self.config.vocab_size]
    
    def _f_skip(self, input_ids: torch.LongTensor):
        """Get the previous token skip."""
        return self.prf[input_ids[- self.config.prefix_length].item()]

    def _f_min(self, input_ids: torch.LongTensor):
        """Get the previous token min."""
        return min(self.prf[input_ids[-1 - i].item()] for i in range(0, self.config.prefix_length))
    
    def get_greenlist_ids(self, input_ids: torch.LongTensor) -> list[int]:
        """Get greenlist ids for the input_ids."""
        return self.window_scheme_map[self.config.window_scheme](input_ids)
    
    def _get_greenlist_ids_left(self, input_ids: torch.LongTensor) -> list[int]:
        """Get greenlist ids for the input_ids via leftHash scheme."""
        self.rng.manual_seed((self.config.hash_key * self._f(input_ids)) % self.config.vocab_size)
        greenlist_size = int(self.config.vocab_size * self.config.gamma)
        vocab_permutation = torch.randperm(self.config.vocab_size, device=input_ids.device, generator=self.rng)
        greenlist_ids = vocab_permutation[:greenlist_size]
        return greenlist_ids
    
    def _get_greenlist_ids_self(self, input_ids: torch.LongTensor) -> list[int]:
        """Get greenlist ids for the input_ids via selfHash scheme."""
        greenlist_size = int(self.config.vocab_size * self.config.gamma)
        greenlist_ids = []
        f_x = self._f(input_ids)
        for k in range(0, self.config.vocab_size):
            h_k = f_x * int(self.prf[k])
            self.rng.manual_seed(h_k % self.config.vocab_size)
            vocab_permutation = torch.randperm(self.config.vocab_size, device=input_ids.device, generator=self.rng)
            temp_greenlist_ids = vocab_permutation[:greenlist_size]
            if k in temp_greenlist_ids:
                greenlist_ids.append(k)
        return greenlist_ids
    
    def _compute_z_score(self, observed_count: int , T: int) -> float: 
        """Compute z-score for the given observed count and total tokens."""
        expected_count = self.config.gamma
        numer = observed_count - expected_count * T 
        denom = sqrt(T * expected_count * (1 - expected_count))  
        z = numer / denom
        return z
    
    def score_sequence(self, input_ids: torch.Tensor) -> tuple[float, list[int]]:
        """Score the input_ids and return z_score and green_token_flags."""
        num_tokens_scored = len(input_ids) - self.config.prefix_length
        if num_tokens_scored < 1:
            raise ValueError(
                (
                    f"Must have at least {1} token to score after "
                    f"the first min_prefix_len={self.config.prefix_length} tokens required by the seeding scheme."
                )
            )

        green_token_count = 0
        green_token_flags = [-1 for _ in range(self.config.prefix_length)]

        for idx in range(self.config.prefix_length, len(input_ids)):
            curr_token = input_ids[idx]
            greenlist_ids = self.get_greenlist_ids(input_ids[:idx])
            if curr_token in greenlist_ids:
                green_token_count += 1
                green_token_flags.append(1)
            else:
                green_token_flags.append(0)
        
        z_score = self._compute_z_score(green_token_count, num_tokens_scored)
        return z_score, green_token_flags


class KGWLogitsProcessor(LogitsProcessor):
    """LogitsProcessor for KGW algorithm, process logits to add watermark."""

    def __init__(self, config: KGWConfig, utils: KGWUtils, *args, **kwargs) -> None:
        """
            Initialize the KGW logits processor.

            Parameters:
                config (KGWConfig): Configuration for the KGW algorithm.
                utils (KGWUtils): Utility class for the KGW algorithm.
        """
        self.config = config
        self.utils = utils

    def _calc_greenlist_mask(self, scores: torch.FloatTensor, greenlist_token_ids: torch.LongTensor) -> torch.BoolTensor:
        """Calculate greenlist mask for the given scores and greenlist token ids."""
        green_tokens_mask = torch.zeros_like(scores)
        for b_idx in range(len(greenlist_token_ids)):
            green_tokens_mask[b_idx][greenlist_token_ids[b_idx]] = 1
        final_mask = green_tokens_mask.bool()
        return final_mask

    def _bias_greenlist_logits(self, scores: torch.Tensor, greenlist_mask: torch.Tensor, greenlist_bias: float) -> torch.Tensor:
        """Bias the scores for the greenlist tokens."""
        scores[greenlist_mask] = scores[greenlist_mask] + greenlist_bias
        return scores

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        """Process logits to add watermark."""
        if input_ids.shape[-1] < self.config.prefix_length:
            return scores

        batched_greenlist_ids = [None for _ in range(input_ids.shape[0])]

        for b_idx in range(input_ids.shape[0]):
            greenlist_ids = self.utils.get_greenlist_ids(input_ids[b_idx])
            batched_greenlist_ids[b_idx] = greenlist_ids

        green_tokens_mask = self._calc_greenlist_mask(scores=scores, greenlist_token_ids=batched_greenlist_ids)

        scores = self._bias_greenlist_logits(scores=scores, greenlist_mask=green_tokens_mask, greenlist_bias=self.config.delta)
        return scores
    

class KGW(BaseWatermark):
    """Top-level class for KGW algorithm."""

    def __init__(self, algorithm_config: str | KGWConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the KGW algorithm.

            Parameters:
                algorithm_config (str | KGWConfig): Path to the algorithm configuration file or KGWConfig instance.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = KGWConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, KGWConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a KGWConfig instance")
            
        self.utils = KGWUtils(self.config)
        self.logits_processor = KGWLogitsProcessor(self.config, self.utils)
    
    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text."""

        # Configure generate_with_watermark
        generate_with_watermark = partial(
            self.config.generation_model.generate,
            logits_processor=LogitsProcessorList([self.logits_processor]), 
            **self.config.gen_kwargs
        )
        
        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)
        # Generate watermarked text
        encoded_watermarked_text = generate_with_watermark(**encoded_prompt)
        # Decode
        watermarked_text = self.config.generation_tokenizer.batch_decode(encoded_watermarked_text, skip_special_tokens=True)[0]
        return watermarked_text
    
    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs):
        """Detect watermark in the text."""

        # Encode the text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)

        # Compute z_score using a utility method
        z_score, _ = self.utils.score_sequence(encoded_text)

        # Determine if the z_score indicates a watermark
        is_watermarked = z_score > self.config.z_threshold

        # Return results based on the return_dict flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": z_score}
        else:
            return (is_watermarked, z_score)
        
    # def get_data_for_visualization(self, text: str, *args, **kwargs) -> tuple[list[str], list[int]]:
    #     """Get data for visualization."""
        
    #     # Encode text
    #     encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)
        
    #     # Compute z-score and highlight values
    #     z_score, highlight_values = self.utils.score_sequence(encoded_text)
        
    #     # decode single tokens
    #     decoded_tokens = []
    #     for token_id in encoded_text:
    #         token = self.config.generation_tokenizer.decode(token_id.item())
    #         decoded_tokens.append(token)
        
    #     return DataForVisualization(decoded_tokens, highlight_values)


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# sir.py
# Description: Implementation of SIR algorithm
# ============================================

import json
import torch
import jieba
import random
import numpy as np
from functools import partial
from ..base import BaseWatermark, BaseConfig
from .transform_model import TransformModel
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from visualize.data_for_visualization import DataForVisualization
from utils.utils import create_directory_for_file, load_config_file
from transformers import LogitsProcessor, LogitsProcessorList, BertTokenizer, BertModel


class SIRConfig(BaseConfig):
    """Config class for SIR algorithm, load config file and initialize parameters."""
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.delta = self.config_dict['delta']
        self.chunk_length = self.config_dict['chunk_length']
        self.scale_dimension = self.config_dict['scale_dimension']
        self.z_threshold = self.config_dict['z_threshold']
        self.transform_model_input_dim = self.config_dict['transform_model_input_dim']
        self.transform_model_name = self.config_dict['transform_model_name']
        self.embedding_model_path = self.config_dict['embedding_model_path']
        self.mapping_name = self.config_dict['mapping_name']
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'SIR'


class SIRUtils:
    """Utility class for SIR algorithm, contains helper functions."""

    def __init__(self, config: SIRConfig, *args, **kwargs) -> None:
        """
            Initialize the SIR utility class.

            Parameters:
                config (SIRConfig): Configuration for the SIR algorithm.
        """
        self.config = config
        self.transform_model = self._get_transform_model(self.config.transform_model_name, config.transform_model_input_dim).to(self.config.device)
        self.embedding_tokenizer = BertTokenizer.from_pretrained(self.config.embedding_model_path)
        self.embedding_model = BertModel.from_pretrained(self.config.embedding_model_path).to(self.config.device)
        self.mapping = self._get_mapping(self.config.mapping_name)

    def get_embedding(self, sentence: str) -> torch.FloatTensor:
        """Get the embedding of the input sentence."""
        input_ids = self.embedding_tokenizer.encode(sentence, return_tensors="pt", max_length=512, truncation="longest_first")
        input_ids = input_ids.to(self.config.device)
        with torch.no_grad():
            output = self.embedding_model(input_ids)
        return output[0][:, 0, :]
    
    def get_text_split(self, sentence: str) -> list[list[str]]:
        """Split the input text into chunks of words."""
        words = list(jieba.cut(sentence))
        non_space_indices = [index for index, word in enumerate(words) if word.strip()]
        words_2d = []
        chunk_start = 0
        for i in range(0, len(non_space_indices), self.config.chunk_length):
            chunk_end = i + self.config.chunk_length
            chunk_end = min(chunk_end, len(non_space_indices))
            chunk_indices = non_space_indices[:chunk_end]
            if chunk_indices:
                chunk = words[chunk_start:chunk_indices[-1] + 1]
                words_2d.append(chunk)
            chunk_start = chunk_indices[-1] + 1
        return words_2d

    def scale_vector(self, v: np.array) -> np.array:
        """Scale the input vector using tanh function."""
        mean = np.mean(v)
        v_minus_mean = v - mean
        v_minus_mean = np.tanh(1000 * v_minus_mean)
        return v_minus_mean
    
    def _get_mapping(self, mapping_name: str) -> list[int]:
        """Get the mapping for the input tokens."""
        input_size = self.config.vocab_size

        # try loading mapping from the provided mapping path
        try:
            with open(mapping_name, 'r') as f:
                mapping = json.load(f)

        # if the file does not exist, create a new mapping and save it to the provided mapping path
        except:
            mapping = [random.randint(0, self.config.scale_dimension - 1) for _ in range(input_size)] 
            create_directory_for_file(mapping_name)
            with open(mapping_name, 'w') as f:
                json.dump(mapping, f, indent=4)
        return mapping
    
    def _get_context_sentence(self, input_ids: torch.LongTensor):
        """Get the context sentence from the input_ids."""
        sentence = self.config.generation_tokenizer.decode(input_ids, skip_special_tokens=True)
        words_2d = self.get_text_split(sentence)
        if len(words_2d) == 0:
            return ''
        if len(words_2d[-1]) == self.config.chunk_length:
            return ''.join([''.join(group) for group in words_2d]).strip()
        else:
            return ''.join([''.join(group) for group in words_2d[:-1]]).strip()
    
    def _get_transform_model(self, model_name: str, input_dim: int) -> TransformModel:
        """Get the transform model from the provided model name."""
        model = TransformModel(input_dim=input_dim)
        model.load_state_dict(torch.load(model_name))
        return model
    
    def get_bias(self, input_ids: torch.LongTensor) -> list[int]:
        """Get the bias for the input_ids."""
        context_sentence = self._get_context_sentence(input_ids)
        context_embedding = self.get_embedding(context_sentence)
        output = self.transform_model(context_embedding).cpu()[0].numpy()
        similarity_array = self.scale_vector(output)[self.mapping]
        return -similarity_array


class SIRLogitsProcessor(LogitsProcessor):
    """Logits processor for SIR algorithm."""

    def __init__(self, config: SIRConfig, utils: SIRUtils, *args, **kwargs):
        """
            Initialize the SIR logits processor.

            Parameters:
                config (SIRConfig): Configuration for the SIR algorithm.
                utils (SIRUtils): Utility class for the SIR algorithm.
        """
        self.config = config
        self.utils = utils
    
    def _bias_logits(self, scores: torch.LongTensor, batched_bias: torch.FloatTensor) -> torch.FloatTensor:
        """Bias the logits using the batched_bias."""
        scores = scores + batched_bias * self.config.delta
        return scores

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        """Process the logits to add watermark."""
        batched_bias = [None for _ in range(input_ids.shape[0])] 

        for b_idx in range(input_ids.shape[0]):
            current_bias = self.utils.get_bias(input_ids[b_idx])
            batched_bias[b_idx] = current_bias

        batched_bias_np = np.array(batched_bias)
        batched_bias = torch.Tensor(batched_bias_np).to(scores.device)

        scores = self._bias_logits(scores=scores, batched_bias=batched_bias)
        return scores


class SIR(BaseWatermark):
    """Top-level class for SIR algorithm."""

    def __init__(self, algorithm_config: str | SIRConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the SIR algorithm.

            Parameters:
                algorithm_config (str): Path to the algorithm configuration file.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = SIRConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, SIRConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a SIRConfig instance")
        
        self.utils = SIRUtils(self.config)
        self.logits_processor = SIRLogitsProcessor(self.config, self.utils)

    def generate_watermarked_text(self, prompt: str, *args, **kwargs):
        """Generate watermarked text."""

        # Configure generate_with_watermark
        generate_with_watermark = partial(
            self.config.generation_model.generate,
            logits_processor=LogitsProcessorList([self.logits_processor]), 
            **self.config.gen_kwargs
        )
        
        # encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)
        # generate watermarked text
        encoded_watermarked_text = generate_with_watermark(**encoded_prompt)
        # decode
        watermarked_text = self.config.generation_tokenizer.batch_decode(encoded_watermarked_text, skip_special_tokens=True)[0]
        return watermarked_text
    
    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs):
        """Detect watermark in the input text."""

        # Split the text into a 2D array of words
        word_2d = self.utils.get_text_split(text)

        # Initialize a list to hold all computed values for similarity
        all_value = []

        # Iterate over each sentence in the split text, skipping the first
        for i in range(1, len(word_2d)):
            # Create the context sentence from all previous text portions
            context_sentence = ''.join([''.join(group) for group in word_2d[:i]]).strip()
            # Current sentence to check against the context
            current_sentence = ''.join(word_2d[i]).strip()

            # Continue if the context sentence is shorter than the required chunk length
            if len(list(jieba.cut(context_sentence))) < self.config.chunk_length:
                continue

            # Get embedding of the context sentence
            context_embedding = self.utils.get_embedding(context_sentence)
            # Transform the embedding using the model, process output
            output = self.utils.transform_model(context_embedding).cpu().detach()[0].numpy()
            # Scale the output vector and map to predefined indices
            similarity_array = self.utils.scale_vector(output)[self.utils.mapping]

            # Encode the current sentence into tokens
            tokens = self.config.generation_tokenizer.encode(current_sentence, return_tensors="pt", add_special_tokens=False)

            # Append negative similarity values for each token in the current sentence
            for index in tokens[0]:
                all_value.append(-float(similarity_array[index]))

        # Calculate the mean of all similarity values
        z_score = np.mean(all_value)

        # Determine if the z_score indicates a watermark
        is_watermarked = z_score > self.config.z_threshold

        # Return results based on the return_dict flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": z_score}
        else:
            return (is_watermarked, z_score)
        
    def get_data_for_visualization(self, text: str, *args, **kwargs):
        """Get data for visualization."""
        
        # Split the text into 2D array of words
        word_2d = self.utils.get_text_split(text)
        highlight_values = []
        decoded_tokens = []

        # Iterate over each sentence in the text
        for i in range(len(word_2d)):
            # Construct the context sentence from the previous sentences
            context_sentence = ' '.join([' '.join(group) for group in word_2d[:i]])
            # Current sentence for tokenization
            current_sentence = ' '.join(word_2d[i])
            # Tokenize the current sentence
            tokens = self.config.generation_tokenizer.encode(current_sentence, return_tensors="pt", add_special_tokens=False)

            # Decode each token and append to the decoded_tokens list
            for token_id in tokens[0]:
                token = self.config.generation_tokenizer.decode(token_id.item())
                decoded_tokens.append(token)

            # If the context sentence is shorter than required, append highlight -1 for each token
            if len(context_sentence.split()) < self.config.chunk_length:
                highlight_values.extend([-1] * len(tokens[0]))
                continue

            # Get the embedding of the context sentence and process it through the model
            context_embedding = self.utils.get_embedding(context_sentence)
            output = self.utils.transform_model(context_embedding).cpu().detach()[0].numpy()

            # Scale the output vector and get similarity values
            similarity_array = self.utils.scale_vector(output)[self.utils.mapping]

            # Append highlight values based on similarity
            for token_index in tokens[0]:
                similarity_value = -float(similarity_array[token_index.item()])
                highlight_values.append(1 if similarity_value > 0 else 0)

        return DataForVisualization(decoded_tokens, highlight_values)



# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ============================================
# network_model.py
# Description: Network model for UPV algorithm
# ============================================

import torch
from torch import nn


class UPVSubNet(nn.Module):
    """Sub-network for UPV algorithm."""

    def __init__(self, input_dim: int, num_layers: int, hidden_dim: int = 64) -> None:
        """
            Initialize the UPV sub-network.

            Parameters:
                input_dim (int): Input dimension.
                num_layers (int): Number of layers.
                hidden_dim (int): Hidden dimension.
        """
        super(UPVSubNet, self).__init__()

        # Define layers
        self.layers = nn.ModuleList()

        # First layer
        self.layers.append(nn.Linear(input_dim, hidden_dim))
        self.layers.append(nn.ReLU())

        # Hidden layers
        for _ in range(num_layers - 1):
            self.layers.append(nn.Linear(hidden_dim, hidden_dim))
            self.layers.append(nn.ReLU())

        # Output layer
        self.layers.append(nn.Linear(hidden_dim, hidden_dim))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        for layer in self.layers:
            x = layer(x)
        return x


class UPVGenerator(nn.Module):
    """Watermark Generator for UPV algorithm."""

    def __init__(self, input_dim: int, window_size: int, num_layers: int = 5, hidden_dim: int = 64):
        """
            Initialize the UPV generator.

            Parameters:
                input_dim (int): Input dimension.
                window_size (int): Window size.
                num_layers (int): Number of layers.
                hidden_dim (int): Hidden dimension.
        """
        super(UPVGenerator, self).__init__()

        # subnet
        self.sub_net = UPVSubNet(input_dim, num_layers, hidden_dim)
        self.window_size = window_size
        self.relu = nn.ReLU()

        # linear layer and sigmoid layer after merging features
        self.combine_layer = nn.Linear(window_size*hidden_dim, hidden_dim)
        self.output_layer = nn.Linear(hidden_dim, 1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for the UPV generator."""

        # x is expected to be of shape (batch_size, window_size, input_dim)
        batch_size = x.shape[0]
        # Reshape x to (-1, input_dim) so it can be passed through the sub_net in one go
        x = x.view(-1, x.shape[-1])
        sub_net_output = self.sub_net(x)
        
        # Reshape sub_net_output back to (batch_size, window_size*hidden_dim)
        sub_net_output = sub_net_output.view(batch_size, -1) 
        combined_features = self.combine_layer(sub_net_output)
        combined_features = self.relu(combined_features)
        output = self.output_layer(combined_features)
        output = self.sigmoid(output)

        return output


class UPVDetector(nn.Module):
    """Watermark Detector for UPV algorithm."""

    def __init__(self, bit_number: int, b_layers: int = 5, input_dim: int = 64, 
                 hidden_dim: int = 128, num_classes: int = 1, num_layers: int = 2):
        """
            Initialize the UPV detector.

            Parameters:
                bit_number (int): Number of bits.
                b_layers (int): Number of layers for the binary classifier.
                input_dim (int): Input dimension.
                hidden_dim (int): Hidden dimension.
                num_classes (int): Number of classes.
                num_layers (int): Number of layers for the classifier.
        """
        super(UPVDetector, self).__init__()
        self.binary_classifier = UPVSubNet(bit_number, b_layers)
        self.classifier = nn.LSTM(input_dim, hidden_dim, num_layers, batch_first=True)
        self.fc_hidden = nn.Linear(hidden_dim, hidden_dim)
        self.fc = nn.Linear(hidden_dim, num_classes)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for the UPV detector."""

        # Reshape the input to process through the binary classifier
        x1 = x.view(-1, x.shape[-1])
        features = self.binary_classifier(x1)
        # Reshape the features to be LSTM compatible
        features = features.view(x.size(0), x.size(1), -1)

        # Process features through the LSTM
        lstm_output, _ = self.classifier(features)
        # Select the last output for classification
        last_lstm_output = lstm_output[:, -1, :]

        # Pass through the hidden fully connected layer
        hidden_output = self.fc_hidden(last_lstm_output)
        hidden_output = self.sigmoid(hidden_output) 

        # Final classification
        final_output = self.fc(hidden_output)
        return self.sigmoid(final_output) 


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Build a unified dictionary from external dictionaries

import random
import argparse
from tqdm import tqdm
from opencc import OpenCC
from itertools import permutations

# set seed
random.seed(0)

S2T = OpenCC('s2t')
T2S = OpenCC('t2s')

def transform(token, append_meta_symbols):
    def capitalize(s):
        return s.capitalize()

    def s2t(s):
        return S2T.convert(s)

    def t2s(s):
        return T2S.convert(s)

    def add_meta_symbols(s):
        # TODO: currently we only consider the sentencepiece meta symbol (U+2581)
        return f"▁{s}"

    # all permutations of the transformations
    transformations = []

    if append_meta_symbols:
        for r in range(1, 5):
            transformations.extend(permutations([capitalize, s2t, t2s, add_meta_symbols], r))
    else:
        for r in range(1, 4):
            transformations.extend(permutations([capitalize, s2t, t2s], r))

    res = [token]
    for t in transformations:
        new_token = token
        for f in t:
            new_token = f(new_token)
        res.append(new_token)

    # deduplicate
    res = list(set(res))
    return res

def augment_dictionary(raw_entries, append_meta_symbols):
    augmented_entries = []
    for src, tgt in tqdm(raw_entries, desc="Augmenting dictionary"):
        src_tokens = transform(src, append_meta_symbols)
        tgt_tokens = transform(tgt, append_meta_symbols)
        for src_token in src_tokens:
            for tgt_token in tgt_tokens:
                if src_token != tgt_token:
                    augmented_entries.append((src_token, tgt_token))

    # deduplicate
    augmented_entries = list(set(augmented_entries))
    return augmented_entries

def main(args):
    # Read data
    raw_entries = [] # list of tuples (src, tgt)
    for d_path in args.dicts:
        with open(d_path, "r") as f:
            for line in f:
                src, tgt = line.strip().split()
                if src != tgt:
                    raw_entries.append((src, tgt))

    # Augment dictionary
    augmented_entries = augment_dictionary(raw_entries, args.append_meta_symbols)

    # Write data
    with open(args.output_file, "w") as f:
        for src, tgt in augmented_entries:
            f.write(f"{src}\t{tgt}\n")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Build a unified dictionary from external dictionaries")
    parser.add_argument("--dicts", type=str, nargs="+", help="multiple external dictionaries")
    parser.add_argument("--output_file", type=str, help="output dictionary")
    parser.add_argument("--append_meta_symbols", action="store_true", help="append meta symbols to the tokens")

    args = parser.parse_args()
    main(args)



# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ==============================================
# pf.py
# Description: Implementation of PF algorithm
# ==============================================

from typing import List
import torch
import numpy as np
from scipy.stats import gamma
from sympy.physics.units import temperature

from ..base import BaseConfig, BaseWatermark
from utils.transformers_config import TransformersConfig

# PF algorithm configuration class
class PFConfig(BaseConfig):
    """Config class for PF algorithm, load config file and initialize parameters."""
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.payload = self.config_dict['payload']
        self.salt_key = self.config_dict['salt_key']
        self.ngram = self.config_dict['ngram']
        self.seed = self.config_dict['seed']
        self.seeding = self.config_dict['seeding']
        self.max_seq_len = self.config_dict['max_seq_len']


    @property
    def algorithm_name(self) -> str:
        """Return algorithm name."""
        return 'PF'


class PFUtils:
    """Utility class for PF algorithm, contains helper functions."""
    def __init__(self, config: PFConfig, *args, **kwargs) -> None:
        """
            Initialize the PF utility class.

            Parameters:
                config (PFConfig): Configuration for the PF algorithm.
        """
        self.config = config
        self.pad_id = config.generation_tokenizer.pad_token_id if config.generation_tokenizer.pad_token_id is not None else config.generation_tokenizer.eos_token_id
        self.eos_id = config.generation_tokenizer.eos_token_id
        self.hashtable = torch.randperm(1000003)
        self.rng = torch.Generator(device=self.config.device)
        self.rng.manual_seed(self.config.seed)

    def hashint(self, integer_tensor: torch.LongTensor) -> torch.LongTensor:
        """Adapted from https://github.com/jwkirchenbauer/lm-watermarking"""
        return self.hashtable[integer_tensor.cpu() % len(self.hashtable)]


    def get_seed_rng(self, input_ids: torch.LongTensor) -> int:
        """get a random seed according to input tokens."""
        if self.config.seeding == 'hash':
            seed = self.config.seed
            for i in input_ids:
                seed = (seed * self.config.salt_key + i.item()) % (2 ** 64 - 1)
        elif self.config.seeding == 'additive':
            seed = self.config.salt_key * torch.sum(input_ids).item()
            seed = self.hashint(seed)
        elif self.config.seeding == 'skip':
            seed = self.config.salt_key * input_ids[0].item()
            seed = self.hashint(seed)
        elif self.config.seeding == 'min':
            seed = self.hashint(self.config.salt_key * input_ids)
            seed = torch.min(seed).item()
        return seed

    def sample_next(
            self,
            logits: torch.FloatTensor,
            ngram_tokens: torch.LongTensor,
            temperature: float,
            top_p: float
    ) -> torch.LongTensor:
        """generate next token from logits."""
        if temperature > 0:
            probs = torch.softmax(logits / temperature, dim=-1)
            probs_sort, probs_idx = torch.sort(probs, dim=-1, descending=True)
            probs_sum = torch.cumsum(probs_sort, dim=-1)
            mask = probs_sum - probs_sort > top_p
            probs_sort[mask] = 0.0
            probs_sort.div_(probs_sort.sum(dim=-1, keepdim=True))
            log_probs = probs_sort.log()

            seed = self.get_seed_rng(ngram_tokens)
            self.rng.manual_seed(seed)

            rs = torch.rand(self.config.vocab_size, generator=self.rng, device=self.rng.device)
            rs = rs.roll(-self.config.payload)
            rs = torch.Tensor(rs).to(probs_sort.device)
            rs = rs[probs_idx]

            log_probs = log_probs - rs.log()


            next_token = torch.argmax(log_probs, dim=-1, keepdim=True)
            next_token = torch.gather(probs_idx, -1, next_token)
        else:
            next_token = torch.argmax(logits, dim=-1)

        return next_token.reshape(-1)



    def score_tok(self, ngram_tokens: List[int], token_id: int) -> torch.Tensor:
        """calculate scores of each token."""
        seed = self.get_seed_rng(torch.tensor(ngram_tokens))
        self.rng.manual_seed(seed)
        rs = torch.rand(self.config.vocab_size, generator=self.rng, device=self.rng.device)
        # avoid log(0)
        rs[rs == 0] = 1e-4
        scores = -rs.log().roll(-token_id)
        return scores

    def get_threshold(self, n_tokens: int, alpha: float = 0.01) -> float:
        """calculate threshold for PF algorithm."""
        if n_tokens <= self.config.ngram:
            return float('inf')

        k = n_tokens - self.config.ngram
        threshold = gamma.ppf(1 - alpha, a=k, scale=1)
        return threshold

    def get_scores_by_t(
            self,
            text: str,
            scoring_method: str = "none",
            ntoks_max: int = None,
            payload_max: int = 0
    ) -> np.array :
        """
        Get score increment for each token in list of texts.
        Args:
            texts: list of texts
            scoring_method:
                'none': score all ngrams
                'v1': only score tokens for which wm window is unique
                'v2': only score unique {wm window+tok} is unique
            ntoks_max: maximum number of tokens
            payload_max: maximum number of messages
        Output:
            score_lists: list of [np array of score increments for every token and payload] for each text
        """
        tokens_id = self.config.generation_tokenizer.encode(text, add_special_tokens=False)
        if ntoks_max is not None:
            tokens_id = tokens_id[:ntoks_max]

        total_len = len(tokens_id)
        start_pos = self.config.ngram + 1
        rts = []
        seen_ntuples = set()

        for cur_pos in range(start_pos, total_len):
            ngram_tokens = tokens_id[cur_pos - self.config.ngram: cur_pos]
            if scoring_method == 'v1':
                tup_for_unique = tuple(ngram_tokens)
                if tup_for_unique in seen_ntuples:
                    continue
                seen_ntuples.add(tup_for_unique)
            elif scoring_method == 'v2':
                tup_for_unique = tuple(ngram_tokens + [tokens_id[cur_pos]])
                if tup_for_unique in seen_ntuples:
                    continue
                seen_ntuples.add(tup_for_unique)

            rt = self.score_tok(ngram_tokens, tokens_id[cur_pos])
            rt = rt[:payload_max + 1]
            rts.append(rt)

        return np.array([rt.cpu().numpy() for rt in rts])

    def get_scores(self,score_lists: np.array) -> float:
        """calculate sum of PF score."""
        if len(score_lists) == 0:
            return 0
        aggregated_score = sum(score_lists)
        return aggregated_score



class PF(BaseWatermark):
    def __init__(self, algorithm_config: str | PFConfig, transformers_config: TransformersConfig | None = None, *args,
                 **kwargs) -> None:
        """
            Initialize the PF algorithm.

            Parameters:
                algorithm_config (str): Path to the algorithm configuration file.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = PFConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, PFConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a SIRConfig instance")

        self.utils = PFUtils(self.config)

    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text using the PF algorithm."""

        prompt_tokens = self.config.generation_tokenizer.encode(prompt, add_special_tokens=False)
        min_prompt_size = len(prompt_tokens)
        total_len = min(self.config.max_seq_len, self.config.gen_kwargs["max_new_tokens"] + min_prompt_size)

        tokens = torch.full((total_len,), self.utils.pad_id).to(self.config.generation_model.device).long()
        tokens[: len(prompt_tokens)] = torch.tensor(prompt_tokens).long()
        input_text_mask = tokens != self.utils.pad_id

        start_pos = min_prompt_size
        prev_pos = 0
        for cur_pos in range(start_pos, total_len):
            outputs = self.config.generation_model.forward(
                tokens[prev_pos:cur_pos].unsqueeze(0), use_cache=True,
                past_key_values=outputs.past_key_values if prev_pos > 0 else None
            )
            ngram_tokens = tokens[cur_pos - self.config.ngram:cur_pos]
            temperature = self.config.gen_kwargs.get("temperature", 0.9)
            top_p = self.config.gen_kwargs.get("top_p", 1.0)
            next_toks = self.utils.sample_next(outputs.logits[:, -1, :], ngram_tokens, temperature, top_p)
            tokens[cur_pos] = torch.where(input_text_mask[cur_pos], tokens[cur_pos], next_toks)
            prev_pos = cur_pos

        tokens = tokens.tolist()

        tokens = tokens[: len(prompt_tokens) + self.config.gen_kwargs["max_new_tokens"]]

        try:
            tokens = tokens[: tokens.index(self.utils.eos_id)]
        except ValueError:
            pass

        return self.config.generation_tokenizer.decode(tokens)

    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs):
        """Detect watermark in the text."""

        scores = self.utils.get_scores_by_t(text)
        score = self.utils.get_scores(scores)
        alpha = self.config.gen_kwargs.get("alpha", 0.01)
        threshold = self.utils.get_threshold(len(scores), alpha)
        result = bool(score > threshold)
        score = float(score)
        if return_dict:
            return {"is_watermarked": result, "score": score}
        else:
            return (result, score)

    def get_data_for_visualize(self, text, *args, **kwargs):
        pass









# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ================================================
# exp_edit.py
# Description: Implementation of EXPEdit algorithm
# ================================================

import torch
import numpy as np
from math import log
from ..base import BaseWatermark, BaseConfig
from .mersenne import MersenneRNG
from utils.utils import load_config_file
from .cython_files.levenshtein import levenshtein
from utils.transformers_config import TransformersConfig
from exceptions.exceptions import AlgorithmNameMismatchError
from visualize.data_for_visualization import DataForVisualization


class EXPEditConfig(BaseConfig):
    """Config class for EXPEdit algorithm, load config file and initialize parameters."""
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.pseudo_length = self.config_dict['pseudo_length']
        self.sequence_length = self.config_dict['sequence_length']
        self.n_runs = self.config_dict['n_runs']
        self.p_threshold = self.config_dict['p_threshold']
        self.key = self.config_dict['key']
        self.top_k = self.config_dict['top_k']
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'EXPEdit'


class EXPEditUtils:
    """Utility class for EXPEdit algorithm, contains helper functions."""

    def __init__(self, config: EXPEditConfig, *args, **kwargs) -> None:
        """
            Initialize the EXPEdit utility class.

            Parameters:
                config (EXPEditConfig): Configuration for the EXPEdit algorithm.
        """
        self.config = config
        self.rng = MersenneRNG(self.config.key)
        self.xi = (torch.tensor([self.rng.rand() for _ in range(self.config.pseudo_length * self.config.vocab_size)])
                   .view(self.config.pseudo_length, self.config.vocab_size))

    def exp_sampling(self, probs: torch.Tensor, u: torch.Tensor) -> torch.Tensor:
        """Sample token using exponential distribution."""

        # If top_k is not specified, use argmax
        if self.config.top_k <= 0:
            return torch.argmax(u ** (1 / probs), axis=1).unsqueeze(-1)
        
        # Ensure top_k is not greater than the vocabulary size
        top_k = min(self.config.top_k, probs.size(-1))
    
        # Get the top_k probabilities and their indices
        top_probs, top_indices = torch.topk(probs, top_k, dim=-1)
    
        # Perform exponential sampling on the top_k probabilities
        sampled_indices = torch.argmax(u.gather(-1, top_indices) ** (1 / top_probs), dim=-1)
    
        # Map back the sampled indices to the original vocabulary indices
        return top_indices.gather(-1, sampled_indices.unsqueeze(-1))
    
    def value_transformation(self, value: float) -> float:
        """Transform value to range [0, 1]."""
        return value/(value + 1)

    def one_run(self, tokens: np.ndarray, xi: np.ndarray) -> tuple:
        """Run one test."""
        k = len(tokens)
        n = len(xi)
        A = np.empty((1,n)) 
        for i in range(1):
            for j in range(n):
                A[i][j] = levenshtein(tokens[i:i+k],xi[(j+np.arange(k))%n],0.0)

        return np.min(A), np.argmin(A)


class EXPEdit(BaseWatermark):
    """Top-level class for the EXPEdit algorithm."""
    def __init__(self, algorithm_config: str | EXPEditConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        if isinstance(algorithm_config, str):
            self.config = EXPEditConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, EXPEditConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a EXPEditConfig instance")
        
        self.utils = EXPEditUtils(self.config)

    def generate_watermarked_text(self, prompt: str, *args, **kwargs):
        """Generate watermarked text."""

        # Encode prompt
        encoded_prompt = self.config.generation_tokenizer.encode(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)

        # Initialize
        shift = torch.randint(self.config.pseudo_length, (1,))
        inputs = encoded_prompt
        attn = torch.ones_like(inputs)
        past = None

        # Generate tokens
        for i in range(self.config.sequence_length):
            with torch.no_grad():
                if past:
                    output = self.config.generation_model(inputs[:,-1:], past_key_values=past, attention_mask=attn)
                else:
                    output = self.config.generation_model(inputs)
            
            # Get probabilities
            probs = torch.nn.functional.softmax(output.logits[:,-1, :self.config.vocab_size], dim=-1).cpu()
            
            # Sample token to add watermark
            token = self.utils.exp_sampling(probs, self.utils.xi[(shift + i) % self.config.pseudo_length,:]).to(self.config.device)
            
            # Update inputs
            inputs = torch.cat([inputs, token], dim=-1)

            # Update past
            past = output.past_key_values

            # Update attention mask
            attn = torch.cat([attn, attn.new_ones((attn.shape[0], 1))], dim=-1)
        
        watermarked_tokens = inputs[0].detach().cpu()
        watermarked_text = self.config.generation_tokenizer.decode(watermarked_tokens, skip_special_tokens=True)

        return watermarked_text

    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs):
        """Detect watermark in the text."""

        encoded_text = self.config.generation_tokenizer.encode(text, return_tensors='pt', add_special_tokens=False).numpy()[0]

        test_result, _ = self.utils.one_run(encoded_text, self.utils.xi.numpy())

        p_val = 0
        
        for i in range(self.config.n_runs):
            xi_alternative = np.random.rand(self.config.pseudo_length, self.config.vocab_size).astype(np.float32)
            null_result, _ = self.utils.one_run(encoded_text, xi_alternative)

            # assuming lower test values indicate presence of watermark
            p_val += null_result <= test_result
            print(f"round: {i + 1}, good: {null_result > test_result}")

        p_val = (p_val + 1.0) / (self.config.n_runs + 1.0)

        # Determine if the computed score exceeds the threshold for watermarking
        is_watermarked = p_val < self.config.p_threshold

        # Return results based on the `return_dict` flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": p_val}
        else:
            return (is_watermarked, p_val)

    def get_data_for_visualization(self, text: str, *args, **kwargs):
        """Get data for visualization."""

        # Encode text
        encoded_text = self.config.generation_tokenizer.encode(text, return_tensors='pt', add_special_tokens=False).numpy()[0]

        # Find best match
        _, index = self.utils.one_run(encoded_text, self.utils.xi.numpy())
        random_numbers = self.utils.xi[(index + np.arange(len(encoded_text))) % len(self.utils.xi)]
        
        highlight_values = []

        # Compute highlight values
        for i in range(0, len(encoded_text)):
            r = random_numbers[i][encoded_text[i]]
            v = log(1/(1 - r))
            v = self.utils.value_transformation(v)
            highlight_values.append(v)

        # Decode each token id to its corresponding string token
        decoded_tokens = []
        for token_id in encoded_text:
            token = self.config.generation_tokenizer.decode(token_id.item())
            decoded_tokens.append(token)
        
        return DataForVisualization(decoded_tokens, highlight_values)


# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# =========================================================================
# AutoWatermark.py
# Description: This is a generic watermark class that will be instantiated
#              as one of the watermark classes of the library when created
#              with the [`AutoWatermark.load`] class method.
# =========================================================================

import torch
import importlib
from typing import List
from watermark.auto_config import AutoConfig

WATERMARK_MAPPING_NAMES={
    'KGW': 'watermark.kgw.KGW',
    'Unigram': 'watermark.unigram.Unigram',
    'SWEET': 'watermark.sweet.SWEET',
    'UPV': 'watermark.upv.UPV',
    'SIR': 'watermark.sir.SIR',
    'XSIR': 'watermark.xsir.XSIR',
    'Unbiased': 'watermark.unbiased.UnbiasedWatermark',
    'DIP': 'watermark.dip.DIP',
    'EWD': 'watermark.ewd.EWD',
    'EXP': 'watermark.exp.EXP',
    'EXPGumbel': 'watermark.exp_gumbel.EXPGumbel',
    'EXPEdit': 'watermark.exp_edit.EXPEdit',
    'ITSEdit': 'watermark.its_edit.ITSEdit',
    'SynthID': 'watermark.synthid.SynthID',
    'TS':'watermark.ts.TS',
    'PF':'watermark.pf.PF',
    'MorphMark':'watermark.morphmark.MorphMark',
}

def watermark_name_from_alg_name(name):
    """Get the watermark class name from the algorithm name."""
    if name in WATERMARK_MAPPING_NAMES:
        return WATERMARK_MAPPING_NAMES[name]
    else:
        raise ValueError(f"Invalid algorithm name: {name}")

class AutoWatermark:
    """
        This is a generic watermark class that will be instantiated as one of the watermark classes of the library when
        created with the [`AutoWatermark.load`] class method.

        This class cannot be instantiated directly using `__init__()` (throws an error).
    """

    def __init__(self):
        raise EnvironmentError(
            "AutoWatermark is designed to be instantiated "
            "using the `AutoWatermark.load(algorithm_name, algorithm_config, transformers_config)` method."
        )

    def load(algorithm_name, algorithm_config=None, transformers_config=None, *args, **kwargs):
        """Load the watermark algorithm instance based on the algorithm name."""
        watermark_name = watermark_name_from_alg_name(algorithm_name)
        module_name, class_name = watermark_name.rsplit('.', 1)
        module = importlib.import_module(module_name)
        watermark_class = getattr(module, class_name)
        watermark_config = AutoConfig.load(algorithm_name, transformers_config, algorithm_config_path=algorithm_config, **kwargs)
        watermark_instance = watermark_class(watermark_config)
        return watermark_instance


vllm_supported_methods = ["UPV", "KGW", "Unigram"]
class AutoWatermarkForVLLM:
    def __init__(self, algorithm_name, algorithm_config, transformers_config):
        if not algorithm_name in vllm_supported_methods:
            raise NotImplementedError(f"vllm integrating currently supports {vllm_supported_methods}, but got {algorithm_name}")
        self.watermark = AutoWatermark.load(algorithm_name=algorithm_name, algorithm_config=algorithm_config, transformers_config=transformers_config)

    def __call__(self, prompt_tokens: List[int], generated_tokens: List[int], scores: torch.FloatTensor) -> torch.Tensor:
        if len(prompt_tokens) == 0:
            return scores
        
        # concencate prompt_tokens and generated_tokens
        input_ids = torch.LongTensor(prompt_tokens + generated_tokens).to(self.watermark.config.device)[None, :]
        scores = scores[None, :]
        assert len(input_ids.shape) == 2, input_ids.shape
        assert len(scores.shape) == 2, scores.shape
        
        scores = self.watermark.logits_processor(input_ids, scores)
        return scores[0, :]

    def get_data_for_visualization(self, text):
        data = self.watermark.get_data_for_visualization(text)
        return data

    def detect_watermark(self, text):
        if type(text) is list:
            return [self.watermark.detect_watermark(_) for _ in text]
        return self.watermark.detect_watermark(text)



import time
import torch
from google import genai
from transformers import AutoModelForCausalLM, AutoTokenizer, LogitsProcessor
import os, json, re
from tqdm import tqdm
from loguru import logger
from functools import partial

def export_openai_key():
    """Export OpenAI API key from config.json to environment variable.
    make sure to have a config.json file with the following structure:
    """
    with open("config.json", "r") as f:
        config = json.load(f)
        os.environ["OPENAI_API_KEY"] = config.get("openai", {}).get("api_key", "")
        assert os.environ["OPENAI_API_KEY"], "OpenAI API key is not set in config.json"
        
def build_prompt(question, retrieved_docs: list[str]) -> str:
    """Build a prompt for RAG-style question answering."""
    prompt = "Here is the set of retrieved documents you will use to answer the question:\n<documents>"
    for doc in retrieved_docs:
        prompt += f"\n{doc}"
    prompt += f"</documents>\n Now, here is the question you need to answer: <question>\n{question}\n</question>"
    return prompt

class huggingface_model:
    def __init__(
        self,
        model_name,
        device="auto",
        watermark = None,
        loaded_model=True,
        login=True,
    ):
        self.model_name = model_name
        self.device = device

        if login:
            # Load the Hugging Face API key from config.json, llama 3 and 4 requires authentication
            import json
            if os.path.exists("config.json") is False:
                logger.info("config.json not found. Please create it with your Hugging Face API key.")
                with open("config.json", "w") as f:
                    json.dump({"huggingface": {"api_key": "YOUR_HUGGINGFACE_API_KEY"}}, f, indent=4)
            else:
                with open("config.json", "r") as f:
                    config = json.load(f)
                api_key = config.get("huggingface", {}).get("api_key")
                if api_key:
                    self._login_huggingface(api_key)

        if loaded_model:
            # Load the model and tokenizer from the Hugging Face Hub
            self.tokenizer = AutoTokenizer.from_pretrained(model_name, dtype=torch.bfloat16)
            logger.info(f"Loaded tokenizer for {model_name} on {device}")
            self.model = AutoModelForCausalLM.from_pretrained(model_name, device_map=device,torch_dtype="auto")
            
        # self.load_watermark(watermark, f"./watermark/{watermark}.json") if watermark else None
        self.watermark = watermark
        self.logit_processors = [watermark.logits_processor] if watermark else list()

    def load_watermark(self, watermark):
        self.watermark = watermark
        self.logit_processors.append(watermark.logits_processor)
    
    def _login_huggingface(self, api_key):
        from huggingface_hub import login
        login(token=api_key)

    def generate(
        self,
        messages,
        do_watermark=False,
        **generation_kwargs,
    ) -> str:
        if do_watermark and not self.logit_processors:
            raise ValueError("Logit processors must be provided for watermarking.")
        
        encoded_chat = self.tokenizer.apply_chat_template(
            messages, 
            return_tensors="pt", 
            add_generation_prompt=True, 
            return_dict=True
        ).to(self.device)

        with torch.no_grad():
            outputs = self.model.generate(
                **encoded_chat,
                **generation_kwargs,
                logits_processor=self.logit_processors if do_watermark else None,
            )
        generated = self.tokenizer.decode(outputs[0][encoded_chat["input_ids"].shape[-1]:]).strip()
        torch.cuda.empty_cache()
        return generated
    
    def inject_watermark_batch(self, doc_lst:list[str]) -> list[str]:
        logger.info(f"inject watermark to {len(doc_lst)} documents")
        wm_docs = list()
        for doc in tqdm(doc_lst):
            messages = [
                {"role": "system", "content": "You are a professional rewriter. Your task is to rewrite the user's input without changing its meaning and length. Do not add any additional information or context, just output the article you re-write."},
                {"role": "user", "content": f"re-write this article:{doc}"},
            ]
            res = self.generate(
                messages,
                do_watermark=True,  # Enable watermarking
                max_length=1024
            )
            wm_docs.append(res)
        return wm_docs

    def detect_watermark(self, text):
        assert self.watermark, "Watermarking is not enabled for this model."
        return self.watermark.detect_watermark(text, return_dict=True)
    

class gpt_oss(huggingface_model):
    def __init__(
        self,
        model_name = "openai/gpt-oss-20b",
        device="cuda",
        watermark = None,
        loaded_model=True,
        login=True,
    ):
        super().__init__(model_name, device, watermark, loaded_model, login)
    
    def extract_final_text(self, decoded: str) -> str:
        # For gpt-oss, it apply harmony chat template, extract final message from the response
        m = re.search(r"<\|start\|>assistant.*?<\|message\|>(.*?)<\|return\|>", decoded, re.S)
        return m.group(1).strip() if m else decoded.strip()

    def generate(
        self,
        messages,
        do_watermark=False,
        **generation_kwargs,
    ) -> str:
        generated = super().generate(messages, do_watermark, **generation_kwargs)
        return self.extract_final_text(generated)
    

class qwen3(huggingface_model):
    def __init__(
        self,
        model_name = "Qwen/Qwen3-8B",
        device="cuda",
        watermark = None,
        loaded_model=True,
        login=True,
        enable_thinking=False
    ):
        super().__init__(model_name, device, watermark, loaded_model, login)
        self.enable_thinking = enable_thinking
    
    def generate(
        self,
        messages,
        do_watermark=False,
        **generation_kwargs,
    ) -> str:
        text = self.tokenizer.apply_chat_template(
            messages,
            tokenize=False,
            add_generation_prompt=True,
            enable_thinking=self.enable_thinking # Switches between thinking and non-thinking modes. Default is True.
        )
        model_inputs = self.tokenizer([text], return_tensors="pt").to(self.model.device)

        with torch.no_grad():
            outputs = self.model.generate(
                **model_inputs,
                **generation_kwargs,
                logits_processor=self.logit_processors if do_watermark else None,
            )
            torch.cuda.empty_cache()
            
        output_ids = outputs[0][len(model_inputs.input_ids[0]):].tolist() 
        

        # parsing thinking content
        try:
            # rindex finding 151668 (</think>)
            index = len(output_ids) - output_ids[::-1].index(151668)
        except ValueError:
            index = 0

        thinking_content = self.tokenizer.decode(output_ids[:index], skip_special_tokens=True).strip("\n")
        content = self.tokenizer.decode(output_ids[index:], skip_special_tokens=True).strip("\n")

        # print("thinking content:", thinking_content)
        # print("content:", content)
        
        return content
    
class gemini(huggingface_model):
    def __init__(self, model_name="gemini-2.5-flash"):
        # We call super().__init__ but disable model loading and HF login,
        # as they are not needed for the API-based Gemini model.
        super().__init__(model_name=model_name, loaded_model=False, login=False)
        
        # Load the Google API key from config.json
        try:
            with open("config.json", "r") as f:
                config = json.load(f)
            api_key = config.get("google", {}).get("api_key")
            if not api_key:
                raise ValueError("API key for Google not found in config.json")
            
            self.client = genai.Client(api_key=config["google"]["api_key"])
            logger.info("Google Generative AI SDK configured successfully.")

        except FileNotFoundError:
            raise FileNotFoundError("config.json not found. Please create it with your Google API key.")
        except json.JSONDecodeError:
            raise ValueError("Could not decode config.json. Please ensure it is valid JSON.")

        # Instantiate the Gemini model
        self.model = partial(self.client.models.generate_content, model=model_name)
        logger.info(f"Gemini model '{self.model_name}' is ready.")

    def generate(
        self,
        messages: list[dict],
        do_watermark: bool = False,
        **generation_kwargs,
    ) -> str:
        if do_watermark:  raise NotImplementedError("Watermarking during generation is not supported by the Gemini API.")
        
        if "max_length" in generation_kwargs:
            generation_kwargs["max_output_tokens"] = generation_kwargs.pop("max_length")
        if "max_new_tokens" in generation_kwargs:
            generation_kwargs["max_output_tokens"] = generation_kwargs.pop("max_new_tokens")
            
        # generation_config = genai.types.GenerationConfig(**generation_kwargs)

        # Convert the message format from 'content' to Gemini's 'parts'
        gemini_messages = []
        for msg in messages:
            # The Gemini API expects the model's role to be 'model', not 'assistant'
            role = "model" if msg["role"] == "assistant" else msg["role"]
            gemini_messages.append({"role": role, "parts": [msg["content"]]})

        try:
            response = self.model(
                contents=str(gemini_messages),
                # config=generation_config
            )
        except Exception as e:
            # tackle with 503 UNAVAILABLE. {'error': {'code': 503, 'message': 'The model is overloaded. Please try again later.', 'status': 'UNAVAILABLE'}}
            # slepp for 30 seconds and retry
            time.sleep(60)
            try:
                response = self.model(
                    contents=str(gemini_messages),
                    # config=generation_config
                )
                return response.text
            except Exception as e:
                logger.error(f"An error occurred during Gemini API call: {e}")
                return f"Error: Failed to generate response. {e}"
        return response.text
    

    
    
if __name__ == "__main__":
    messages = [
        {"role": "system", "content": "You are a helpful assistant. Answer concisely."},
        {"role": "user", "content": "Who is the president of Japan?"},
    ]
    
    
    gen_model = qwen3(enable_thinking=True)
    generated = gen_model.generate(messages, max_new_tokens=10240)
    print(generated)
    
    



# Copyright 2024 THU-BPM MarkLLM.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# ================================================
# unigram.py
# Description: Implementation of Unigram algorithm
# ================================================

from typing import Union
import torch
import hashlib
import numpy as np
from math import sqrt
from functools import partial
from ..base import BaseWatermark, BaseConfig
# from utils.utils import load_config_file
from watermark.transformers_config import TransformersConfig
# from exceptions.exceptions import AlgorithmNameMismatchError
from transformers import LogitsProcessor, LogitsProcessorList
# from visualize.data_for_visualization import DataForVisualization


class UnigramConfig(BaseConfig):
    """Config class for Unigram algorithm, load config file and initialize parameters."""

    # def __init__(self, algorithm_config: str, transformers_config: TransformersConfig, *args, **kwargs) -> None:
    #     """
    #         Initialize the Unigram configuration.

    #         Parameters:
    #             algorithm_config (str): Path to the algorithm configuration file.
    #             transformers_config (TransformersConfig): Configuration for the transformers model.
    #     """
    #     if algorithm_config is None:
    #         config_dict = load_config_file('config/Unigram.json')
    #     else:
    #         config_dict = load_config_file(algorithm_config)
    #     if config_dict['algorithm_name'] != 'Unigram':
    #         raise AlgorithmNameMismatchError('Unigram', config_dict['algorithm_name'])
        
    #     self.gamma = config_dict['gamma']
    #     self.delta = config_dict['delta']
    #     self.hash_key = config_dict['hash_key']
    #     self.z_threshold = config_dict['z_threshold']

    #     self.generation_model = transformers_config.model
    #     self.generation_tokenizer = transformers_config.tokenizer
    #     self.vocab_size = transformers_config.vocab_size
    #     self.device = transformers_config.device
    #     self.gen_kwargs = transformers_config.gen_kwargs
    
    def initialize_parameters(self) -> None:
        """Initialize algorithm-specific parameters."""
        self.gamma = self.config_dict['gamma']
        self.delta = self.config_dict['delta']
        self.hash_key = self.config_dict['hash_key']
        self.z_threshold = self.config_dict['z_threshold']
    
    @property
    def algorithm_name(self) -> str:
        """Return the algorithm name."""
        return 'Unigram'


class UnigramUtils:
    """Utility class for Unigram algorithm, contains helper functions."""

    def __init__(self, config: UnigramConfig, *args, **kwargs) -> None:
        """
            Initialize the Unigram utility class.

            Parameters:
                config (UnigramConfig): Configuration for the Unigram algorithm.
        """
        self.config = config
        self.mask = np.array([True] * int(self.config.gamma * self.config.vocab_size) + 
                             [False] * (self.config.vocab_size - int(self.config.gamma * self.config.vocab_size)))
        self.rng = np.random.default_rng(self._hash_fn(self.config.hash_key))
        self.rng.shuffle(self.mask)
        
    @staticmethod
    def _hash_fn(x: int) -> int:
        """hash function to generate random seed, solution from https://stackoverflow.com/questions/67219691/python-hash-function-that-returns-32-or-64-bits"""
        x = np.int64(x)
        return int.from_bytes(hashlib.sha256(x).digest()[:4], 'little')
    
    def _compute_z_score(self, observed_count: int, T: int) -> float:
        """Compute z-score for the given observed count and total tokens."""
        expected_count = self.config.gamma
        numer = observed_count - expected_count * T 
        denom = sqrt(T * expected_count * (1 - expected_count))  
        z = numer / denom
        return z

    def score_sequence(self, input_ids: torch.Tensor) -> tuple:
        """Score the input_ids and return z_score and green_token_flags."""
        num_tokens_scored = len(input_ids)
        green_token_count = 0
        green_token_flags = []
        for idx in range(0, len(input_ids)):
            curr_token = input_ids[idx]
            if self.mask[curr_token] == True:
                green_token_count += 1
                green_token_flags.append(1)
            else:
                green_token_flags.append(0)

        z_score = self._compute_z_score(green_token_count, num_tokens_scored)
        return z_score, green_token_flags


class UnigramLogitsProcessor(LogitsProcessor):
    """Logits processor for Unigram algorithm."""

    def __init__(self, config: UnigramConfig, utils: UnigramUtils, *args, **kwargs):
        """
            Initialize the Unigram logits processor.

            Parameters:
                config (UnigramConfig): Configuration for the Unigram algorithm.
                utils (UnigramUtils): Utility class for the Unigram algorithm.
        """
        self.config = config
        self.utils = utils
        self.green_list_mask = torch.tensor(self.utils.mask, dtype=torch.float32)

    def _bias_greenlist_logits(self, scores: torch.Tensor, greenlist_mask: torch.Tensor, greenlist_bias: float) -> torch.Tensor:
        """Bias the logits for the greenlist tokens."""
        scores[greenlist_mask] = scores[greenlist_mask] + greenlist_bias
        return scores

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor) -> torch.FloatTensor:
        """Process the logits and add watermark."""
        greenlist_mask = torch.zeros_like(scores)
        for i in range(input_ids.shape[0]):
            greenlist_mask[i] = self.green_list_mask
        scores = self._bias_greenlist_logits(scores=scores, greenlist_mask=greenlist_mask.bool(), greenlist_bias=self.config.delta)
        return scores


class Unigram(BaseWatermark):
    """Top-level class of Unigram algorithm"""

    def __init__(self, algorithm_config: str | UnigramConfig, transformers_config: TransformersConfig | None = None, *args, **kwargs) -> None:
        """
            Initialize the Unigram algorithm.

            Parameters:
                algorithm_config (str | UnigramConfig): Path to the algorithm configuration file or UnigramConfig instance.
                transformers_config (TransformersConfig): Configuration for the transformers model.
        """
        if isinstance(algorithm_config, str):
            self.config = UnigramConfig(algorithm_config, transformers_config)
        elif isinstance(algorithm_config, UnigramConfig):
            self.config = algorithm_config
        else:
            raise TypeError("algorithm_config must be either a path string or a UnigramConfig instance")
        
        self.utils = UnigramUtils(self.config)
        self.logits_processor = UnigramLogitsProcessor(self.config, self.utils)

    def generate_watermarked_text(self, prompt: str, *args, **kwargs) -> str:
        """Generate watermarked text using the prompt."""

        # Configure generate_with_watermark
        generate_with_watermark = partial(
            self.config.generation_model.generate,
            logits_processor=LogitsProcessorList([self.logits_processor]), 
            **self.config.gen_kwargs
        )
        
        # encode prompt
        encoded_prompt = self.config.generation_tokenizer(prompt, return_tensors="pt", add_special_tokens=True).to(self.config.device)
        # generate watermarked text
        encoded_watermarked_text = generate_with_watermark(**encoded_prompt)
        # decode
        watermarked_text = self.config.generation_tokenizer.batch_decode(encoded_watermarked_text, skip_special_tokens=True)[0]
        return watermarked_text
        
    def detect_watermark(self, text: str, return_dict: bool = True, *args, **kwargs) -> Union[tuple, dict]:
        """Detect watermark in the given text."""

        # encode text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)
        
        # compute z_score
        z_score, _ = self.utils.score_sequence(encoded_text)

        # Determine if the z_score indicates a watermark
        is_watermarked = z_score > self.config.z_threshold

        # Return results based on the return_dict flag
        if return_dict:
            return {"is_watermarked": is_watermarked, "score": z_score}
        else:
            return (is_watermarked, z_score)
    
    def get_data_for_visualization(self, text: str, *args, **kwargs) -> tuple[list[str], list[int]]:
        """Get data for visualization."""
        
        # encode text
        encoded_text = self.config.generation_tokenizer(text, return_tensors="pt", add_special_tokens=False)["input_ids"][0].to(self.config.device)
        
        # compute z-score and highlight values
        z_score, highlight_values = self.utils.score_sequence(encoded_text)
        
        # decode single tokens
        decoded_tokens = []
        for token_id in encoded_text:
            token = self.config.generation_tokenizer.decode(token_id.item())
            decoded_tokens.append(token)
        
        return DataForVisualization(decoded_tokens, highlight_values)